@@ -5,64 +5,106 @@ Build all of the packages in a given directory.
 """
 
 import argparse
-from functools import total_ordering
+import dataclasses
+import hashlib
 import json
-from pathlib import Path
-from queue import Queue, PriorityQueue
+import os
 import shutil
 import subprocess
 import sys
-from threading import Thread, Lock
-from time import sleep, perf_counter
-from typing import Dict, Set, Optional, List, Any
-import os
+from collections.abc import Iterable
+from functools import total_ordering
+from graphlib import TopologicalSorter
+from pathlib import Path
+from queue import PriorityQueue, Queue
+from threading import Lock, Thread
+from time import perf_counter, sleep
+from typing import Any
 
 from . import common
-from .io import parse_package_config
-from .common import UNVENDORED_STDLIB_MODULES
 from .buildpkg import needs_rebuild
+from .common import UNVENDORED_STDLIB_MODULES, find_matching_wheels
+from .io import parse_package_config
 
 
+class BuildError(Exception):
+    def __init__(self, returncode: int) -> None:
+        self.returncode = returncode
+        super().__init__()
+
+
+@total_ordering
+@dataclasses.dataclass(eq=False, repr=False)
 class BasePackage:
     pkgdir: Path
     name: str
     version: str
-    meta: dict
+    disabled: bool
+    cpython_dynlib: bool
+    meta: dict[str, Any]
     library: bool
     shared_library: bool
-    dependencies: List[str]
-    unbuilt_dependencies: Set[str]
-    dependents: Set[str]
-    unvendored_tests: Optional[bool] = None
+    dependencies: list[str]
+    unbuilt_dependencies: set[str]
+    dependents: set[str]
+    unvendored_tests: Path | None = None
+    file_name: str | None = None
+    install_dir: str = "site"
+    _queue_idx: int | None = None
 
     # We use this in the priority queue, which pops off the smallest element.
     # So we want the smallest element to have the largest number of dependents
-    def __lt__(self, other) -> bool:
+    def __lt__(self, other: Any) -> bool:
         return len(self.dependents) > len(other.dependents)
 
-    def __eq__(self, other) -> bool:
+    def __eq__(self, other: Any) -> bool:
         return len(self.dependents) == len(other.dependents)
 
+    def __repr__(self) -> str:
+        return f"{type(self).__name__}({self.name})"
 
-@total_ordering
+    def needs_rebuild(self) -> bool:
+        return needs_rebuild(
+            self.pkgdir, self.pkgdir / "build", self.meta.get("source", {})
+        )
+
+    def build(self, outputdir: Path, args: Any) -> None:
+        raise NotImplementedError()
+
+    def wheel_path(self) -> Path:
+        raise NotImplementedError()
+
+    def tests_path(self) -> Path | None:
+        return None
+
+
+@dataclasses.dataclass
 class StdLibPackage(BasePackage):
     def __init__(self, pkgdir: Path):
         self.pkgdir = pkgdir
         self.meta = {}
         self.name = pkgdir.stem
         self.version = "1.0"
+        self.disabled = False
         self.library = False
         self.shared_library = False
         self.dependencies = []
         self.unbuilt_dependencies = set()
         self.dependents = set()
+        self.install_dir = "lib"
 
-    def build(self, outputdir: Path, args) -> None:
+    def build(self, outputdir: Path, args: Any) -> None:
         # All build / packaging steps are already done in the main Makefile
         return
 
+    def wheel_path(self) -> Path:
+        raise RuntimeError("StdLibPackage has no wheel")
 
-@total_ordering
+    def tests_path(self) -> Path | None:
+        return None
+
+
+@dataclasses.dataclass
 class Package(BasePackage):
     def __init__(self, pkgdir: Path):
         self.pkgdir = pkgdir
@@ -74,6 +116,8 @@ class Package(BasePackage):
         self.meta = parse_package_config(pkgpath)
         self.name = self.meta["package"]["name"]
         self.version = self.meta["package"]["version"]
+        self.disabled = self.meta["package"].get("_disabled", False)
+        self.cpython_dynlib = self.meta["package"].get("_cpython_dynlib", False)
         self.meta["build"] = self.meta.get("build", {})
         self.meta["requirements"] = self.meta.get("requirements", {})
 
@@ -86,91 +130,81 @@ class Package(BasePackage):
         self.unbuilt_dependencies = set(self.dependencies)
         self.dependents = set()
 
-    def build(self, outputdir: Path, args) -> None:
-        with open(self.pkgdir / "build.log.tmp", "w") as f:
-            p = subprocess.run(
-                [
-                    sys.executable,
-                    "-m",
-                    "pyodide_build",
-                    "buildpkg",
-                    str(self.pkgdir / "meta.yaml"),
-                    "--cflags",
-                    args.cflags,
-                    "--cxxflags",
-                    args.cxxflags,
-                    "--ldflags",
-                    args.ldflags,
-                    "--target-install-dir",
-                    args.target_install_dir,
-                    "--host-install-dir",
-                    args.host_install_dir,
-                    # Either this package has been updated and this doesn't
-                    # matter, or this package is dependent on a package that has
-                    # been updated and should be rebuilt even though its own
-                    # files haven't been updated.
-                    "--force-rebuild",
-                ],
-                check=False,
-                stdout=f,
-                stderr=subprocess.STDOUT,
+    def wheel_path(self) -> Path:
+        dist_dir = self.pkgdir / "dist"
+        wheels = list(find_matching_wheels(dist_dir.glob("*.whl")))
+        if len(wheels) != 1:
+            raise RuntimeError(
+                f"Unexpected number of wheels {len(wheels)} when building {self.name}"
             )
+        return wheels[0]
 
-        # Don't overwrite build log if we didn't build the file.
-        # If the file didn't need to be rebuilt, the log will have exactly two lines.
-        rebuilt = True
-        with open(self.pkgdir / "build.log.tmp", "r") as f:
-            try:
-                next(f)
-                next(f)
-                next(f)
-            except StopIteration:
-                rebuilt = False
+    def tests_path(self) -> Path | None:
+        tests = list((self.pkgdir / "dist").glob("*-tests.tar"))
+        assert len(tests) <= 1
+        if tests:
+            return tests[0]
+        return None
 
-        if rebuilt:
-            shutil.move(self.pkgdir / "build.log.tmp", self.pkgdir / "build.log")  # type: ignore
-        else:
-            (self.pkgdir / "build.log.tmp").unlink()
+    def build(self, outputdir: Path, args: Any) -> None:
 
-        if args.log_dir and (self.pkgdir / "build.log").exists():
+        p = subprocess.run(
+            [
+                sys.executable,
+                "-m",
+                "pyodide_build",
+                "buildpkg",
+                str(self.pkgdir / "meta.yaml"),
+                "--cflags",
+                args.cflags,
+                "--cxxflags",
+                args.cxxflags,
+                "--ldflags",
+                args.ldflags,
+                "--target-install-dir",
+                args.target_install_dir,
+                "--host-install-dir",
+                args.host_install_dir,
+                # Either this package has been updated and this doesn't
+                # matter, or this package is dependent on a package that has
+                # been updated and should be rebuilt even though its own
+                # files haven't been updated.
+                "--force-rebuild",
+            ],
+            check=False,
+            stdout=subprocess.DEVNULL,
+            stderr=subprocess.DEVNULL,
+        )
+
+        log_dir = Path(args.log_dir).resolve() if args.log_dir else None
+        if log_dir and (self.pkgdir / "build.log").exists():
+            log_dir.mkdir(exist_ok=True, parents=True)
             shutil.copy(
-                self.pkgdir / "build.log", Path(args.log_dir) / f"{self.name}.log"
+                self.pkgdir / "build.log",
+                log_dir / f"{self.name}.log",
             )
 
-        try:
-            p.check_returncode()
-        except subprocess.CalledProcessError:
+        if p.returncode != 0:
             print(f"Error building {self.name}. Printing build logs.")
 
-            with open(self.pkgdir / "build.log", "r") as f:
+            with open(self.pkgdir / "build.log") as f:
                 shutil.copyfileobj(f, sys.stdout)
 
-            raise
+            print("ERROR: cancelling buildall")
+            raise BuildError(p.returncode)
 
         if self.library:
             return
-        shutil.copyfile(
-            self.pkgdir / "build" / (self.name + ".data"),
-            outputdir / (self.name + ".data"),
-        )
-        shutil.copyfile(
-            self.pkgdir / "build" / (self.name + ".js"),
-            outputdir / (self.name + ".js"),
-        )
-        if (self.pkgdir / "build" / (self.name + "-tests.data")).exists():
-            shutil.copyfile(
-                self.pkgdir / "build" / (self.name + "-tests.data"),
-                outputdir / (self.name + "-tests.data"),
-            )
-            shutil.copyfile(
-                self.pkgdir / "build" / (self.name + "-tests.js"),
-                outputdir / (self.name + "-tests.js"),
-            )
+        if self.shared_library:
+            file_path = Path(self.pkgdir / f"{self.name}-{self.version}.zip")
+            shutil.copy(file_path, outputdir)
+            file_path.unlink()
+            return
 
 
 def generate_dependency_graph(
-    packages_dir: Path, packages: Set[str]
-) -> Dict[str, BasePackage]:
+    packages_dir: Path, packages: set[str]
+) -> dict[str, BasePackage]:
     """This generates a dependency graph for listed packages.
 
     A node in the graph is a BasePackage object defined above, which maintains
@@ -190,51 +224,94 @@ def generate_dependency_graph(
     Returns:
      - pkg_map: dictionary mapping package names to BasePackage objects
     """
-
-    pkg_map: Dict[str, BasePackage] = {}
+    pkg: BasePackage
+    pkgname: str
+    pkg_map: dict[str, BasePackage] = {}
 
     if "*" in packages:
         packages.discard("*")
         packages.update(
-            str(x) for x in packages_dir.iterdir() if (x / "meta.yaml").is_file()
+            str(x.name) for x in packages_dir.iterdir() if (x / "meta.yaml").is_file()
         )
 
     no_numpy_dependents = "no-numpy-dependents" in packages
     if no_numpy_dependents:
         packages.discard("no-numpy-dependents")
 
+    disabled_packages = set()
+    for pkgname in list(packages):
+        if pkgname.startswith("!"):
+            packages.discard(pkgname)
+            disabled_packages.add(pkgname[1:])
+
+    # Record which packages were requested. We need this information because
+    # some packages are reachable from the initial set but are only reachable
+    # via a disabled dependency.
+    # Example: scikit-learn needs joblib & scipy, scipy needs numpy but numpy disabled.
+    # We don't want to build joblib.
+    requested = set(packages)
+
+    # Create dependency graph.
+    # On first pass add all dependencies regardless of whether
+    # disabled since it might happen because of a transitive dependency
+    graph = {}
     while packages:
         pkgname = packages.pop()
 
-        pkg: BasePackage
         if pkgname in UNVENDORED_STDLIB_MODULES:
             pkg = StdLibPackage(packages_dir / pkgname)
         else:
             pkg = Package(packages_dir / pkgname)
-        if no_numpy_dependents and "numpy" in pkg.dependencies:
-            continue
-        pkg_map[pkg.name] = pkg
-
+        pkg_map[pkgname] = pkg
+        graph[pkgname] = pkg.dependencies
         for dep in pkg.dependencies:
             if pkg_map.get(dep) is None:
                 packages.add(dep)
 
-    # Compute dependents
-    for pkg in pkg_map.values():
+    # Traverse in build order (dependencies first then dependents)
+    # Mark a package as disabled if they've either been explicitly disabled
+    # or if any of its transitive dependencies were marked disabled.
+    for pkgname in TopologicalSorter(graph).static_order():
+        pkg = pkg_map[pkgname]
+        if pkgname in disabled_packages:
+            pkg.disabled = True
+            continue
+        if no_numpy_dependents and "numpy" in pkg.dependencies:
+            pkg.disabled = True
+            continue
+        for dep in pkg.dependencies:
+            if pkg_map[dep].disabled:
+                pkg.disabled = True
+                break
+
+    # Now traverse in reverse build order (dependents first then their
+    # dependencies).
+    # Locate the subset of packages that are transitive dependencies of packages
+    # that are requested and not disabled.
+    for pkgname in reversed(list(TopologicalSorter(graph).static_order())):
+        pkg = pkg_map[pkgname]
+        if pkg.disabled:
+            requested.discard(pkgname)
+            continue
+
+        if pkgname not in requested:
+            continue
+
+        requested.update(pkg.dependencies)
         for dep in pkg.dependencies:
             pkg_map[dep].dependents.add(pkg.name)
 
-    return pkg_map
+    return {name: pkg_map[name] for name in requested}
 
 
-def job_priority(pkg: BasePackage):
+def job_priority(pkg: BasePackage) -> int:
     if pkg.name == "numpy":
         return 0
     else:
         return 1
 
 
-def print_with_progress_line(str, progress_line):
+def print_with_progress_line(str: str, progress_line: str | None) -> None:
     if not sys.stdout.isatty():
         print(str)
         return
@@ -245,13 +322,13 @@ def print_with_progress_line(str, progress_line):
         print(progress_line, end="\r")
 
 
-def get_progress_line(package_set):
+def get_progress_line(package_set: dict[str, None]) -> str | None:
     if not package_set:
         return None
-    return f"In progress: " + ", ".join(package_set.keys())
+    return "In progress: " + ", ".join(package_set.keys())
 
 
-def format_name_list(l: List[str]) -> str:
+def format_name_list(l: list[str]) -> str:
     """
     >>> format_name_list(["regex"])
     'regex'
@@ -269,8 +346,8 @@ def format_name_list(l: List[str]) -> str:
 
 
 def mark_package_needs_build(
-    pkg_map: Dict[str, BasePackage], pkg: BasePackage, needs_build: Set[str]
-):
+    pkg_map: dict[str, BasePackage], pkg: BasePackage, needs_build: set[str]
+) -> None:
     """
     Helper for generate_needs_build_set. Modifies needs_build in place.
     Recursively add pkg and all of its dependencies to needs_build.
@@ -284,7 +361,7 @@ def mark_package_needs_build(
         mark_package_needs_build(pkg_map, pkg_map[dep], needs_build)
 
 
-def generate_needs_build_set(pkg_map):
+def generate_needs_build_set(pkg_map: dict[str, BasePackage]) -> set[str]:
     """
     Generate the set of packages that need to be rebuilt.
 
@@ -293,15 +370,17 @@ def generate_needs_build_set(pkg_map):
        according to needs_rebuild, and
     2. packages which depend on case 1 packages.
     """
-    needs_build = set()
+    needs_build: set[str] = set()
     for pkg in pkg_map.values():
         # Otherwise, rebuild packages that have been updated and their dependents.
-        if needs_rebuild(pkg.pkgdir, pkg.pkgdir / "build", pkg.meta):
+        if pkg.needs_rebuild():
             mark_package_needs_build(pkg_map, pkg, needs_build)
     return needs_build
 
 
-def build_from_graph(pkg_map: Dict[str, BasePackage], outputdir: Path, args) -> None:
+def build_from_graph(
+    pkg_map: dict[str, BasePackage], outputdir: Path, args: argparse.Namespace
+) -> None:
     """
     This builds packages in pkg_map in parallel, building at most args.n_jobs
     packages at once.
@@ -321,7 +400,7 @@ def build_from_graph(pkg_map: Dict[str, BasePackage], outputdir: Path, args) ->
 
     # Insert packages into build_queue. We *must* do this after counting
     # dependents, because the ordering ought not to change after insertion.
-    build_queue: PriorityQueue = PriorityQueue()
+    build_queue: PriorityQueue[tuple[int, BasePackage]] = PriorityQueue()
 
     if args.force_rebuild:
         # If "force_rebuild" is set, just rebuild everything
@@ -352,12 +431,13 @@ def build_from_graph(pkg_map: Dict[str, BasePackage], outputdir: Path, args) ->
         if len(pkg.unbuilt_dependencies) == 0:
             build_queue.put((job_priority(pkg), pkg))
 
-    built_queue: Queue = Queue()
+    built_queue: Queue[BasePackage | Exception] = Queue()
     thread_lock = Lock()
     queue_idx = 1
-    package_set = {}
+    # Using dict keys for insertion order preservation
+    package_set: dict[str, None] = {}
 
-    def builder(n):
+    def builder(n: int) -> None:
         nonlocal queue_idx
         while True:
             pkg = build_queue.get()[1]
@@ -392,9 +472,15 @@ def build_from_graph(pkg_map: Dict[str, BasePackage], outputdir: Path, args) ->
 
     num_built = len(already_built)
     while num_built < len(pkg_map):
-        pkg = built_queue.get()
-        if isinstance(pkg, Exception):
-            raise pkg
+        match built_queue.get():
+            case BuildError() as err:
+                raise SystemExit(err.returncode)
+            case Exception() as err:
+                raise err
+            case a_package:
+                # MyPy should understand that this is a BasePackage
+                assert not isinstance(a_package, Exception)
+                pkg = a_package
 
         num_built += 1
 
@@ -404,51 +490,50 @@ def build_from_graph(pkg_map: Dict[str, BasePackage], outputdir: Path, args) ->
             if len(dependent.unbuilt_dependencies) == 0:
                 build_queue.put((job_priority(dependent), dependent))
 
-    for name in list(pkg_map):
-        if (outputdir / (name + "-tests.js")).exists():
-            pkg_map[name].unvendored_tests = True
-
     print(
         "\n===================================================\n"
         f"built all packages in {perf_counter() - t0:.2f} s"
     )
 
 
-def generate_packages_json(pkg_map: Dict[str, BasePackage]) -> Dict:
-    """Generate the package.json file"""
-    # Build package.json data.
-    package_data: Dict[str, Dict[str, Any]] = {
-        "info": {"arch": "wasm32", "platform": "Emscripten-1.0"},
-        "packages": {},
-    }
+def _generate_package_hash(full_path: Path) -> str:
+    sha256_hash = hashlib.sha256()
+    with open(full_path, "rb") as f:
+        while chunk := f.read(4096):
+            sha256_hash.update(chunk)
+    return sha256_hash.hexdigest()
 
+
+def generate_packagedata(
+    output_dir: Path, pkg_map: dict[str, BasePackage]
+) -> dict[str, Any]:
     libraries = [pkg.name for pkg in pkg_map.values() if pkg.library]
-
-    # unvendored stdlib modules
-    for name in UNVENDORED_STDLIB_MODULES:
-        pkg_entry: Dict[str, Any] = {
-            "name": name,
-            "version": "1.0",
-            "depends": [],
-            "imports": [name],
-        }
-        package_data["packages"][name.lower()] = pkg_entry
-
+    packages: dict[str, Any] = {}
     for name, pkg in pkg_map.items():
-        if pkg.library:
+        if not pkg.file_name:
             continue
-        pkg_entry = {"name": name, "version": pkg.version}
+        if not Path(output_dir, pkg.file_name).exists():
+            continue
+        pkg_entry: Any = {
+            "name": name,
+            "version": pkg.version,
+            "file_name": pkg.file_name,
+            "install_dir": pkg.install_dir,
+            "sha256": _generate_package_hash(Path(output_dir, pkg.file_name)),
+        }
         if pkg.shared_library:
             pkg_entry["shared_library"] = True
+            pkg_entry["install_dir"] = "lib" if pkg.cpython_dynlib else "dynlib"
+
         pkg_entry["depends"] = [
             x.lower() for x in pkg.dependencies if x not in libraries
         ]
         pkg_entry["imports"] = pkg.meta.get("test", {}).get("imports", [name])
 
-        package_data["packages"][name.lower()] = pkg_entry
+        packages[name.lower()] = pkg_entry
 
         if pkg.unvendored_tests:
-            package_data["packages"][name.lower()]["unvendored_tests"] = True
+            packages[name.lower()]["unvendored_tests"] = True
 
             # Create the test package if necessary
             pkg_entry = {
@@ -456,34 +541,92 @@ def generate_packages_json(pkg_map: Dict[str, BasePackage]) -> Dict:
                 "version": pkg.version,
                 "depends": [name.lower()],
                 "imports": [],
+                "file_name": pkg.unvendored_tests.name,
+                "install_dir": pkg.install_dir,
+                "sha256": _generate_package_hash(
+                    Path(output_dir, pkg.unvendored_tests.name)
+                ),
             }
-            package_data["packages"][name.lower() + "-tests"] = pkg_entry
+            packages[name.lower() + "-tests"] = pkg_entry
 
     # Workaround for circular dependency between soupsieve and beautifulsoup4
     # TODO: FIXME!!
-    if "soupsieve" in package_data["packages"]:
-        package_data["packages"]["soupsieve"]["depends"].append("beautifulsoup4")
+    if "soupsieve" in packages:
+        packages["soupsieve"]["depends"].append("beautifulsoup4")
 
-    # re-order packages by name
-    package_data["packages"] = dict(sorted(package_data["packages"].items()))
-
-    return package_data
+    # sort packages by name
+    packages = dict(sorted(packages.items()))
+    return packages
 
 
-def build_packages(packages_dir: Path, outputdir: Path, args) -> None:
+def generate_repodata(
+    output_dir: Path, pkg_map: dict[str, BasePackage]
+) -> dict[str, dict[str, Any]]:
+    """Generate the package.json file"""
+
+    import sys
+
+    sys.path.append(str(common.get_pyodide_root() / "src/py"))
+    from pyodide import __version__
+
+    # Build package.json data.
+    [platform, _, arch] = common.platform().rpartition("_")
+    info = {
+        "arch": arch,
+        "platform": platform,
+        "version": __version__,
+        "python": sys.version.partition(" ")[0],
+    }
+    packages = generate_packagedata(output_dir, pkg_map)
+    return dict(info=info, packages=packages)
+
+
+def copy_packages_to_dist_dir(
+    packages: Iterable[BasePackage], output_dir: Path
+) -> None:
+    for pkg in packages:
+        try:
+            shutil.copy(pkg.wheel_path(), output_dir)
+        except RuntimeError:
+            pass
+
+        test_path = pkg.tests_path()
+        if test_path:
+            shutil.copy(test_path, output_dir)
+
+
+def build_packages(
+    packages_dir: Path, output_dir: Path, args: argparse.Namespace
+) -> None:
     packages = common._parse_package_subset(args.only)
 
     pkg_map = generate_dependency_graph(packages_dir, packages)
 
-    build_from_graph(pkg_map, outputdir, args)
+    build_from_graph(pkg_map, output_dir, args)
+    for pkg in pkg_map.values():
+        if pkg.library:
+            continue
+        if isinstance(pkg, StdLibPackage):
+            pkg.file_name = pkg.name + ".tar"
+            continue
+        if pkg.needs_rebuild():
+            continue
+        if pkg.shared_library:
+            pkg.file_name = f"{pkg.name}-{pkg.version}.zip"
+            continue
+        assert isinstance(pkg, Package)
+        pkg.file_name = pkg.wheel_path().name
+        pkg.unvendored_tests = pkg.tests_path()
 
-    package_data = generate_packages_json(pkg_map)
+    copy_packages_to_dist_dir(pkg_map.values(), output_dir)
+    package_data = generate_repodata(output_dir, pkg_map)
 
-    with open(outputdir / "packages.json", "w") as fd:
+    with open(output_dir / "repodata.json", "w") as fd:
         json.dump(package_data, fd)
+        fd.write("\n")
 
 
-def make_parser(parser):
+def make_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
     parser.description = (
         "Build all the packages in a given directory\n\n"
         "Unless the --only option is provided\n\n"
@@ -494,12 +637,14 @@ def make_parser(parser):
         "dir",
         type=str,
         nargs=1,
+        default="packages",
         help="Input directory containing a tree of package definitions",
     )
     parser.add_argument(
         "output",
         type=str,
         nargs=1,
+        default="dist",
         help="Output directory in which to put all built packages",
     )
     parser.add_argument(
@@ -569,9 +714,10 @@ def make_parser(parser):
     return parser
 
 
-def main(args):
+def main(args: argparse.Namespace) -> None:
     packages_dir = Path(args.dir[0]).resolve()
     outputdir = Path(args.output[0]).resolve()
+    outputdir.mkdir(exist_ok=True)
     if args.cflags is None:
         args.cflags = common.get_make_flag("SIDE_MODULE_CFLAGS")
     if args.cxxflags is None:
